#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

#define DEBUG
#define DEV_PG      "/dev/pglove0"
#define DEV_MIDI    "/dev/midi00"
#define CC_VOL      0x07
#define CC_CUTOFF   0x4a
#define CC_RES      0x2a
#define MSG_CC      0xb0
#define MSG_NOTEON  0x90
#define MSG_NOTEOFF 0x80
#define delay(x)    usleep(x)

typedef struct
{
  unsigned char status;
  signed char x;
  signed char y;
  signed char z;
  unsigned char rot;
  unsigned char fingers;
  unsigned char keys;
  unsigned char dummy7;
  unsigned char dummy8;
  unsigned char dummy9;
  unsigned char dummy10;
  unsigned char dummy11;
} glove_data;

int main(int argc, char** argv)
{
  int pgfd=0;
  int midifd=0;
  char buf[12]={0};
  unsigned char msg[3];
  glove_data* data=(glove_data*)buf;

  if((pgfd=open(DEV_PG, O_RDONLY)) < 0)
  {
    printf("can't open DEV_PG, %d\n", pgfd);
    exit(1);
  }

  if((midifd=open(DEV_MIDI, O_RDWR)) < 0)
  {
    printf("can't open DEV_MIDI, %d\n", midifd);
    exit(1);
  }

  printf("pgfd=%d\n",pgfd);
  printf("midifd=%d\n", midifd);

  while(1)
  {
    read(pgfd, buf, sizeof(buf));
#ifdef DEBUG
    printf("[%2.2x %+4i %+4i %+4i %2.2x] ", data->status, data->x, data->y, data->z, data->fingers);
#endif

    msg[0]=MSG_CC;
    msg[1]=CC_CUTOFF;
    msg[2]=data->x & 0x7f;
    write(midifd, msg, sizeof(msg));
#ifdef DEBUG
    printf("[%2.2x %2.2x %2.2x] ", msg[0], msg[1], msg[2]);
#endif

    msg[0]=MSG_CC;
    msg[1]=CC_VOL;
    msg[2]=data->fingers & 0x7f;
    write(midifd, msg, sizeof(msg));
#ifdef DEBUG
    printf("[%2.2x %2.2x %2.2x]\n", msg[0], msg[1], msg[2]);
#endif

    delay(10100);
  }

  close(pgfd);
  close(midifd);
  return 0;
}
