/*
linux hi-res powerglove driver 1.4, 8/13/2002
rewrite/port to linux 2.4 by sean whalen <lucidity@gmx.net>
original code by manfredo@opal.cs.tu-berlin.de and greg alt <galt@eskimo.com>

this driver works with the parallel interface from byte magazine, july 1990.
interface diagram and updates can be found at http://chocobospore.org
*/

#include "linux/types.h"
#include "linux/fs.h"
#include "linux/kernel.h"
#include "linux/module.h"
#include "linux/sched.h"
#include "linux/signal.h"
#include "linux/wrapper.h"
#include "linux/interrupt.h"
#include "linux/delay.h"
#include "asm/irq.h"
#include "asm/io.h"
#include "asm/delay.h"

#define GDATA       0x10    /* PG data in */
#define GLATCH      0x02    /* PG latch out */
#define GCLOCK      0x01    /* PG clock out */
#define GCLOLAT     0x03    /* clock + latch */
#define D2BYTES     96      /* delay between 2 bytes   96 us */
#define D2BITS      22      /* delay between 2 bits    22 us */
#define D2SLOW      14720   /* 5 slow bytes in packet  14720 us */
#define OUTPORT     0x378
#define INPORT      0x379
#define C0L0()      outb(0, OUTPORT)        /* clock 0 latch 0 */
#define C0L1()      outb(GLATCH, OUTPORT)   /* clock 0 latch 1 */
#define C1L0()      outb(GCLOCK, OUTPORT)   /* clock 1 latch 0 */
#define C1L1()      outb(GCLOLAT, OUTPORT)  /* clock 1 latch 1 */
#define readbit(x)  x<<=1; x+=(inb(INPORT) & GDATA) >> 4;
#define MODNAME     "pglove"
#define delay(x)    udelay(x)

static int major=0;
int initialized=0;

ssize_t pglove_read(struct file*, char*, size_t, loff_t*);
int pglove_open(struct inode*, struct file*);
int pglove_close(struct inode*, struct file*);
void get_glove_data(unsigned char *);
unsigned char get_byte(void);
void init_hires(void);

struct file_operations pglove_fops={
  read:     pglove_read,
  open:     pglove_open,
  release:  pglove_close,
};

int init_module(void)
{
  major=register_chrdev(major, MODNAME, &pglove_fops);

  if(major<0)
  {
    printk("error, can't get major: %d\n", major);
    return major;
  }

  printk("registered module, major=%d\n", major);
  return 0;
}

int cleanup_module(void)
{
  unregister_chrdev(major, MODNAME);
  return 0;
}

ssize_t pglove_read(struct file* file, char* buffer, size_t size, loff_t* offset)
{
  if(!initialized)
  {
    init_hires();
  }

  get_glove_data(buffer);
  return size;
}

int pglove_open(struct inode* inode, struct file* file)
{
  MOD_INC_USE_COUNT;
  initialized=0;
  return 0;
}

int pglove_close(struct inode* inode, struct file* file)
{
  MOD_DEC_USE_COUNT;
  initialized=0;
  return 0;
}

void get_glove_data(unsigned char *buf)
{
  *(buf)=get_byte(); delay(D2BYTES);
  *(buf+1)=get_byte(); delay(D2BYTES);
  *(buf+2)=get_byte(); delay(D2BYTES);
  *(buf+3)=get_byte(); delay(D2BYTES);
  *(buf+4)=get_byte(); delay(D2BYTES);
  *(buf+5)=get_byte(); delay(D2BYTES);
  *(buf+6)=get_byte(); delay(D2BYTES);
  *(buf+7)=get_byte(); delay(D2SLOW);
  *(buf+8)=get_byte(); delay(D2SLOW);
  *(buf+10)=get_byte(); delay(D2SLOW);
  *(buf+11)=get_byte(); delay(D2SLOW);
}

unsigned char get_byte(void)
{
  unsigned char data=0;

  C1L0(); C1L1(); delay(5);
  C1L0();
  readbit(data); C0L0(); C1L0();
  readbit(data); C0L0(); C1L0();
  readbit(data); C0L0(); C1L0();
  readbit(data); C0L0(); C1L0();
  readbit(data); C0L0(); C1L0();
  readbit(data); C0L0(); C1L0();
  readbit(data); C0L0(); C1L0();
  readbit(data); C0L0(); C1L0();

  return data;
}

void init_hires(void)
{
  unsigned char dummy=0;

  //discard first four bits
  C1L0(); C1L1(); delay(5); C1L0();
  readbit(dummy); C0L0(); C1L0();
  readbit(dummy); C0L0(); C1L0();
  readbit(dummy); C0L0(); C1L0();
  readbit(dummy); C0L0(); C1L0();
 
  C1L0(); delay(7212);
  C1L1(); delay(2260);
 
  //byte1: 0x06
  C1L0();
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C1L1();
  C0L1(); C1L1(); delay(D2BITS);
  C0L1(); C1L1(); delay(D2BITS);
  C1L0();
  C0L0(); C1L0(); delay(D2BYTES);

  //byte2: 0xc1
  C1L1();
  C0L1(); C1L1(); delay(D2BITS);
  C0L1(); C1L1(); delay(D2BITS);
  C1L0();
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C1L1();
  C0L1(); C1L1(); delay(D2BYTES);
 
  //byte3: 0x08
  C1L0();
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C1L1();
  C0L1(); C1L1(); delay(D2BITS);
  C1L0();
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BYTES);
 
  //byte4: 0x00
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BYTES);
 
  //byte5: 0x02
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C1L1();
  C0L1(); C1L1(); delay(D2BITS);
  C1L0();
  C0L0(); C1L0(); delay(D2BYTES);
 
  //byte6: 0xff
  C1L1();
  C0L1(); C1L1(); delay(D2BITS);
  C0L1(); C1L1(); delay(D2BITS);
  C0L1(); C1L1(); delay(D2BITS);
  C0L1(); C1L1(); delay(D2BITS);
  C0L1(); C1L1(); delay(D2BITS);
  C0L1(); C1L1(); delay(D2BITS);
  C0L1(); C1L1(); delay(D2BITS);
  C0L1(); C1L1(); delay(D2BYTES);
 
  //byte7: 0x01 or 0x00 (faster)
  C1L0();
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C0L0(); C1L0(); delay(D2BITS);
  C1L1();
  C0L1(); C1L1();  delay(892);
 
  C1L0();
  delay(10000); //let glove settle
  delay(10000);
  delay(10000);
  delay(10000);

  initialized=1;
}

MODULE_LICENSE("GPL");
